@@ -1,27 +1,20 @@
 /* Create, modify, and extract from archives.
-   Copyright (C) 2005, 2007 Red Hat, Inc.
+   Copyright (C) 2005-2012, 2016, 2017 Red Hat, Inc.
+   This file is part of elfutils.
    Written by Ulrich Drepper <drepper@redhat.com>, 2005.
 
-   Red Hat elfutils is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by the
-   Free Software Foundation; version 2 of the License.
+   This file is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 3 of the License, or
+   (at your option) any later version.
 
-   Red Hat elfutils is distributed in the hope that it will be useful, but
+   elfutils is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License along
-   with Red Hat elfutils; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301 USA.
-
-   Red Hat elfutils is an included package of the Open Invention Network.
-   An included package of the Open Invention Network is a package for which
-   Open Invention Network licensees cross-license their patents.  No patent
-   license is granted, either expressly or impliedly, by designation as an
-   included package.  Should you wish to participate in the Open Invention
-   Network licensing program, please visit www.openinventionnetwork.com
-   <http://www.openinventionnetwork.com>.  */
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -29,13 +22,10 @@
 
 #include <argp.h>
 #include <assert.h>
-#include <error.h>
 #include <fcntl.h>
 #include <gelf.h>
-#include <libintl.h>
 #include <limits.h>
 #include <locale.h>
-#include <mcheck.h>
 #include <search.h>
 #include <stdbool.h>
 #include <stdlib.h>
@@ -45,17 +35,18 @@
 #include <time.h>
 #include <unistd.h>
 #include <sys/mman.h>
-#include <sys/statfs.h>
+#include <sys/stat.h>
 #include <sys/time.h>
 
 #include <system.h>
+#include <printversion.h>
 
 #include "arlib.h"
 
 
 /* Name and version of program.  */
-static void print_version (FILE *stream, struct argp_state *state);
-void (*argp_program_version_hook) (FILE *, struct argp_state *) = print_version;
+ARGP_PROGRAM_VERSION_HOOK_DEF = print_version;
+
 /* Prototypes for local functions.  */
 static int do_oper_extract (int oper, const char *arfname, char **argv,
 			    int argc, long int instance);
@@ -66,13 +57,13 @@ static int do_oper_insert (int oper, const char *arfname, char **argv,
 
 
 /* Bug report address.  */
-const char *argp_program_bug_address = PACKAGE_BUGREPORT;
+ARGP_PROGRAM_BUG_ADDRESS_DEF = PACKAGE_BUGREPORT;
 
 
 /* Definitions of arguments for argp functions.  */
 static const struct argp_option options[] =
 {
-  { NULL, 0, NULL, 0, N_("Commands:"), 0 },
+  { NULL, 0, NULL, 0, N_("Commands:"), 1 },
   { NULL, 'd', NULL, 0, N_("Delete files from archive."), 0 },
   { NULL, 'm', NULL, 0, N_("Move files in archive."), 0 },
   { NULL, 'p', NULL, 0, N_("Print files in archive."), 0 },
@@ -82,7 +73,7 @@ static const struct argp_option options[] =
   { NULL, 't', NULL, 0, N_("Display content of archive."), 0 },
   { NULL, 'x', NULL, 0, N_("Extract files from archive."), 0 },
 
-  { NULL, 0, NULL, 0, N_("Command Modifiers:"), 0 },
+  { NULL, 0, NULL, 0, N_("Command Modifiers:"), 2 },
   { NULL, 'o', NULL, 0, N_("Preserve original dates."), 0 },
   { NULL, 'N', NULL, 0, N_("Use instance [COUNT] of name."), 0 },
   { NULL, 'C', NULL, 0,
@@ -114,7 +105,7 @@ static error_t parse_opt (int key, char *arg, struct argp_state *state);
 /* Data structure to communicate with argp functions.  */
 static struct argp argp =
 {
-  options, parse_opt, args_doc, doc, NULL, NULL, NULL
+  options, parse_opt, args_doc, doc, arlib_argp_children, NULL, NULL
 };
 
 
@@ -147,9 +138,6 @@ static enum { ipos_none, ipos_before, ipos_after } ipos;
 int
 main (int argc, char *argv[])
 {
-  /* Make memory leak detection possible.  */
-  mtrace ();
-
   /* We use no threads here which can interfere with handling a stream.  */
   (void) __fsetlocking (stdin, FSETLOCKING_BYCALLER);
   (void) __fsetlocking (stdout, FSETLOCKING_BYCALLER);
@@ -187,14 +175,16 @@ main (int argc, char *argv[])
   if (ipos != ipos_none)
     {
       /* Only valid for certain operations.  */
-      if (operation == oper_extract && operation == oper_delete)
-	error (1, 0, gettext ("\
+      if (operation != oper_move && operation != oper_replace)
+	error (1, 0, _("\
 'a', 'b', and 'i' are only allowed with the 'm' and 'r' options"));
 
       if (remaining == argc)
 	{
-	  error (0, 0, gettext ("MEMBER parameter required"));
-	  argp_help (&argp, stderr, ARGP_HELP_SEE, AR);
+	  error (0, 0, _("\
+MEMBER parameter required for 'a', 'b', and 'i' modifiers"));
+	  argp_help (&argp, stderr, ARGP_HELP_USAGE | ARGP_HELP_SEE,
+		     program_invocation_short_name);
 	  exit (EXIT_FAILURE);
 	}
 
@@ -206,14 +196,15 @@ main (int argc, char *argv[])
   if (instance_specifed)
     {
       /* Only valid for certain operations.  */
-      if (operation == oper_extract && operation == oper_delete)
-	error (1, 0, gettext ("\
+      if (operation != oper_extract && operation != oper_delete)
+	error (1, 0, _("\
 'N' is only meaningful with the 'x' and 'd' options"));
 
       if (remaining == argc)
 	{
-	  error (0, 0, gettext ("COUNT parameter required"));
-	  argp_help (&argp, stderr, ARGP_HELP_SEE, AR);
+	  error (0, 0, _("COUNT parameter required"));
+	  argp_help (&argp, stderr, ARGP_HELP_SEE,
+		     program_invocation_short_name);
 	  exit (EXIT_FAILURE);
 	}
 
@@ -223,21 +214,21 @@ main (int argc, char *argv[])
 	   && errno == ERANGE)
 	  || instance <= 0
 	  || *endp != '\0')
-	error (1, 0, gettext ("invalid COUNT parameter %s"), argv[remaining]);
+	error (1, 0, _("invalid COUNT parameter %s"), argv[remaining]);
 
       ++remaining;
     }
 
   if ((dont_replace_existing || allow_truncate_fname)
       && unlikely (operation != oper_extract))
-    error (1, 0, gettext ("'%' is only meaningful with the 'x' option"),
+    error (1, 0, _("'%c' is only meaningful with the 'x' option"),
 	   dont_replace_existing ? 'C' : 'T');
 
   /* There must at least be one more parameter specifying the archive.   */
   if (remaining == argc)
     {
-      error (0, 0, gettext ("Archive name required"));
-      argp_help (&argp, stderr, ARGP_HELP_SEE, AR);
+      error (0, 0, _("archive name required"));
+      argp_help (&argp, stderr, ARGP_HELP_SEE, program_invocation_short_name);
       exit (EXIT_FAILURE);
     }
 
@@ -248,6 +239,13 @@ main (int argc, char *argv[])
   int status;
   switch (operation)
     {
+    case oper_none:
+      error (0, 0, _("command option required"));
+      argp_help (&argp, stderr, ARGP_HELP_STD_ERR,
+		 program_invocation_short_name);
+      status = 1;
+      break;
+
     case oper_list:
     case oper_print:
       status = do_oper_extract (operation, arfname, argv, argc, -1);
@@ -277,20 +275,6 @@ main (int argc, char *argv[])
 }
 
 
-/* Print the version information.  */
-static void
-print_version (FILE *stream, struct argp_state *state __attribute__ ((unused)))
-{
-  fprintf (stream, "ar (%s) %s\n", PACKAGE_NAME, PACKAGE_VERSION);
-  fprintf (stream, gettext ("\
-Copyright (C) %s Red Hat, Inc.\n\
-This is free software; see the source for copying conditions.  There is NO\n\
-warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
-"), "2007");
-  fprintf (stream, gettext ("Written by %s.\n"), "Ulrich Drepper");
-}
-
-
 /* Handle program arguments.  */
 static error_t
 parse_opt (int key, char *arg __attribute__ ((unused)),
@@ -307,8 +291,9 @@ parse_opt (int key, char *arg __attribute__ ((unused)),
     case 'x':
       if (operation != oper_none)
 	{
-	  error (0, 0, gettext ("More than one operation specified"));
-	  argp_help (&argp, stderr, ARGP_HELP_SEE, AR);
+	  error (0, 0, _("More than one operation specified"));
+	  argp_help (&argp, stderr, ARGP_HELP_SEE,
+		     program_invocation_short_name);
 	  exit (EXIT_FAILURE);
 	}
 
@@ -375,6 +360,10 @@ parse_opt (int key, char *arg __attribute__ ((unused)),
       allow_truncate_fname = true;
       break;
 
+    case 'u':
+      update_newer = true;
+      break;
+
     case 'v':
       verbose = true;
       break;
@@ -396,8 +385,8 @@ open_archive (const char *arfname, int flags, int mode, Elf **elf,
       if (miss_allowed)
 	return -1;
 
-      error (EXIT_FAILURE, errno, gettext ("cannot open archive '%s'"),
-	     arfname);
+      error_exit (errno, _("cannot open archive '%s'"),
+		  arfname);
     }
 
   if (elf != NULL)
@@ -406,16 +395,16 @@ open_archive (const char *arfname, int flags, int mode, Elf **elf,
 
       *elf = elf_begin (fd, cmd, NULL);
       if (*elf == NULL)
-	error (EXIT_FAILURE, 0, gettext ("cannot open archive '%s': %s"),
-	       arfname, elf_errmsg (-1));
+	error_exit (0, _("cannot open archive '%s': %s"),
+		    arfname, elf_errmsg (-1));
 
       if (flags == O_RDONLY && elf_kind (*elf) != ELF_K_AR)
-	error (EXIT_FAILURE, 0, gettext ("%s: not an archive file"), arfname);
+	error_exit (0, _("%s: not an archive file"), arfname);
     }
 
   if (st != NULL && fstat (fd, st) != 0)
-    error (EXIT_FAILURE, errno, gettext ("cannot stat archive '%s'"),
-	   arfname);
+    error_exit (errno, _("cannot stat archive '%s'"),
+		arfname);
 
   return fd;
 }
@@ -426,7 +415,7 @@ not_found (int argc, char *argv[argc], bool found[argc])
 {
   for (int i = 0; i < argc; ++i)
     if (!found[i])
-      printf (gettext ("no entry %s in archive\n"), argv[i]);
+      printf (_("no entry %s in archive\n"), argv[i]);
 }
 
 
@@ -446,17 +435,30 @@ copy_content (Elf *elf, int newfd, off_t off, size_t n)
   return write_retry (newfd, rawfile + off, n) != (ssize_t) n;
 }
 
+static inline bool
+should_truncate_fname (size_t *name_max)
+{
+  if (errno == ENAMETOOLONG && allow_truncate_fname)
+    {
+      if (*name_max == 0)
+	{
+	  long int len = pathconf (".", _PC_NAME_MAX);
+	  if (len > 0)
+	    *name_max = len;
+	}
+      return *name_max != 0;
+    }
+  return false;
+}
 
 static int
 do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 		 long int instance)
 {
-  bool found[argc];
+  bool found[argc > 0 ? argc : 1];
   memset (found, '\0', sizeof (found));
 
-  struct statfs f;
-  f.f_namelen = 0;
-
+  size_t name_max = 0;
   off_t index_off = -1;
   size_t index_size = 0;
   off_t cur_off = SARMAG;
@@ -466,14 +468,13 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
   int fd = open_archive (arfname, O_RDONLY, 0, &elf, NULL, false);
 
   if (hcreate (2 * argc) == 0)
-    error (EXIT_FAILURE, errno, gettext ("cannot create hash table"));
+    error_exit (errno, _("cannot create hash table"));
 
   for (int cnt = 0; cnt < argc; ++cnt)
     {
       ENTRY entry = { .key = argv[cnt], .data = &argv[cnt] };
       if (hsearch (entry, ENTER) == NULL)
-	error (EXIT_FAILURE, errno,
-	       gettext ("cannot insert into hash table"));
+	error_exit (errno, _("cannot insert into hash table"));
     }
 
   struct stat st;
@@ -481,7 +482,7 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
     {
       if (fstat (fd, &st) != 0)
 	{
-	  error (0, errno, gettext ("cannot stat '%s'"), arfname);
+	  error (0, errno, _("cannot stat '%s'"), arfname);
 	  close (fd);
 	  return 1;
 	}
@@ -516,7 +517,7 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 	  ENTRY entry;
 	  entry.key = arhdr->ar_name;
 	  ENTRY *res = hsearch (entry, FIND);
-	  if (res != NULL && (instance < 0 || instance-- == 0)
+	  if (res != NULL && (instance < 0 || --instance == 0)
 	      && !found[(char **) res->data - argv])
 	    found[(char **) res->data - argv] = do_extract = true;
 	}
@@ -536,8 +537,14 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 	      else if (oper == oper_list)
 		{
 		  char datestr[100];
-		  strftime (datestr, sizeof (datestr), "%b %e %H:%M %Y",
-			    localtime (&arhdr->ar_date));
+		  struct tm *tp = localtime (&arhdr->ar_date);
+		  if (tp == NULL)
+		    {
+		      time_t time = 0;
+		      tp = localtime (&time);
+		    }
+
+		  strftime (datestr, sizeof (datestr), "%b %e %H:%M %Y", tp);
 
 		  printf ("%c%c%c%c%c%c%c%c%c %u/%u %6ju %s %s\n",
 			  (arhdr->ar_mode & S_IRUSR) ? 'r' : '-',
@@ -577,7 +584,7 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 	  char *data = elf_rawfile (subelf, &nleft);
 	  if (data == NULL)
 	    {
-	      error (0, 0, gettext ("cannot read content of %s: %s"),
+	      error (0, 0, _("cannot read content of %s: %s"),
 		     arhdr->ar_name, elf_errmsg (-1));
 	      status = 1;
 	      goto next;
@@ -606,22 +613,21 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 		    {
 		      int printlen = INT_MAX;
 
-		      if (errno == ENAMETOOLONG && allow_truncate_fname
-			  && (f.f_namelen != 0 || statfs (".", &f) == 0))
+		      if (should_truncate_fname (&name_max))
 			{
 			  /* Try to truncate the name.  First find out by how
 			     much.  */
-			  printlen = f.f_namelen;
-			  char truncfname[f.f_namelen + 1];
+			  printlen = name_max;
+			  char truncfname[name_max + 1];
 			  *((char *) mempcpy (truncfname, arhdr->ar_name,
-					      f.f_namelen)) = '\0';
+					      name_max)) = '\0';
 
 			  xfd = open (truncfname, flags, 0600);
 			}
 
 		      if (xfd == -1)
 			{
-			  error (0, errno, gettext ("cannot open %.*s"),
+			  error (0, errno, _("cannot open %.*s"),
 				 (int) printlen, arhdr->ar_name);
 			  status = 1;
 			  goto next;
@@ -643,7 +649,7 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 
 	  if (unlikely (n == -1))
 	    {
-	      error (0, errno, gettext ("failed to write %s"), arhdr->ar_name);
+	      error (0, errno, _("failed to write %s"), arhdr->ar_name);
 	      status = 1;
 	      unlink (tempfname);
 	      close (xfd);
@@ -655,23 +661,23 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 	      /* Fix up the mode.  */
 	      if (unlikely (fchmod (xfd, arhdr->ar_mode) != 0))
 		{
-		  error (0, errno, gettext ("cannot change mode of %s"),
+		  error (0, errno, _("cannot change mode of %s"),
 			 arhdr->ar_name);
 		  status = 0;
 		}
 
 	      if (preserve_dates)
 		{
-		  struct timeval tv[2];
+		  struct timespec tv[2];
 		  tv[0].tv_sec = arhdr->ar_date;
-		  tv[0].tv_usec = 0;
+		  tv[0].tv_nsec = 0;
 		  tv[1].tv_sec = arhdr->ar_date;
-		  tv[1].tv_usec = 0;
+		  tv[1].tv_nsec = 0;
 
-		  if (unlikely (futimes (xfd, tv) != 0))
+		  if (unlikely (futimens (xfd, tv) != 0))
 		    {
 		      error (0, errno,
-			     gettext ("cannot change modification time of %s"),
+			     _("cannot change modification time of %s"),
 			     arhdr->ar_name);
 		      status = 1;
 		    }
@@ -696,15 +702,14 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 		    {
 		      int printlen = INT_MAX;
 
-		      if (errno == ENAMETOOLONG && allow_truncate_fname
-			  && (f.f_namelen != 0 || statfs (".", &f) == 0))
+		      if (should_truncate_fname (&name_max))
 			{
 			  /* Try to truncate the name.  First find out by how
 			     much.  */
-			  printlen = f.f_namelen;
-			  char truncfname[f.f_namelen + 1];
+			  printlen = name_max;
+			  char truncfname[name_max + 1];
 			  *((char *) mempcpy (truncfname, arhdr->ar_name,
-					      f.f_namelen)) = '\0';
+					      name_max)) = '\0';
 
 			  if (dont_replace_existing)
 			    {
@@ -718,7 +723,7 @@ do_oper_extract (int oper, const char *arfname, char **argv, int argc,
 
 		      if (r != 0)
 			{
-			  error (0, errno, gettext ("\
+			  error (0, errno, _("\
 cannot rename temporary file to %.*s"),
 				 printlen, arhdr->ar_name);
 			  unlink (tempfname);
@@ -754,7 +759,7 @@ cannot rename temporary file to %.*s"),
 	  if (unlikely (newfd == -1))
 	    {
 	    nonew:
-	      error (0, errno, gettext ("cannot create new file"));
+	      error (0, errno, _("cannot create new file"));
 	      status = 1;
 	    }
 	  else
@@ -780,26 +785,30 @@ cannot rename temporary file to %.*s"),
 	      else
 		rest_off = SARMAG;
 
-	      if ((symtab.symsnamelen != 0
+	      if (symtab.symsnamelen != 0
 		   && ((write_retry (newfd, symtab.symsoff,
 				     symtab.symsofflen)
 			!= (ssize_t) symtab.symsofflen)
 		       || (write_retry (newfd, symtab.symsname,
 					symtab.symsnamelen)
 			   != (ssize_t) symtab.symsnamelen)))
-		  /* Even if the original file had content before the
-		     symbol table, we write it in the correct order.  */
-		  || (index_off != SARMAG
-		      && copy_content (elf, newfd, SARMAG, index_off - SARMAG))
-		  || copy_content (elf, newfd, rest_off, st.st_size - rest_off)
-		  /* Set the mode of the new file to the same values the
-		     original file has.  */
-		  || fchmod (newfd, st.st_mode & ALLPERMS) != 0
-		  /* Never complain about fchown failing.  */
-		  || (({asm ("" :: "r" (fchown (newfd, st.st_uid,
-						st.st_gid))); }),
-		      close (newfd) != 0)
-		  || (newfd = -1, rename (tmpfname, arfname) != 0))
+		goto nonew_unlink;
+	      /* Even if the original file had content before the
+		 symbol table, we write it in the correct order.  */
+	      if ((index_off != SARMAG
+		   && copy_content (elf, newfd, SARMAG, index_off - SARMAG))
+		  || copy_content (elf, newfd, rest_off, st.st_size - rest_off))
+		goto nonew_unlink;
+
+	      /* Never complain about fchown failing.  */
+	      if (fchown (newfd, st.st_uid, st.st_gid) != 0) { ; }
+	      /* Set the mode of the new file to the same values the
+		 original file has.  */
+	      if (fchmod (newfd, st.st_mode & ALLPERMS) != 0
+		  || close (newfd) != 0)
+		goto nonew_unlink;
+	      newfd = -1;
+	      if (rename (tmpfname, arfname) != 0)
 		goto nonew_unlink;
 	    }
 	}
@@ -837,7 +846,10 @@ write_member (struct armem *memb, off_t *startp, off_t *lenp, Elf *elf,
 	      off_t end_off, int newfd)
 {
   struct ar_hdr arhdr;
-  char tmpbuf[sizeof (arhdr.ar_name) + 1];
+  /* The ar_name is not actually zero terminated, but we need that for
+     snprintf.  Also if the name is too long, then the string starts
+     with '/' plus an index off number (decimal).  */
+  char tmpbuf[sizeof (arhdr.ar_name) + 2];
 
   bool changed_header = memb->long_name_off != -1;
   if (changed_header)
@@ -899,7 +911,7 @@ do_oper_delete (const char *arfname, char **argv, int argc,
 		long int instance)
 {
   bool *found = alloca (sizeof (bool) * argc);
-  memset (found, '\0', sizeof (found));
+  memset (found, '\0', sizeof (bool) * argc);
 
   /* List of the files we keep.  */
   struct armem *to_copy = NULL;
@@ -910,14 +922,13 @@ do_oper_delete (const char *arfname, char **argv, int argc,
   int fd = open_archive (arfname, O_RDONLY, 0, &elf, &st, false);
 
   if (hcreate (2 * argc) == 0)
-    error (EXIT_FAILURE, errno, gettext ("cannot create hash table"));
+    error_exit (errno, _("cannot create hash table"));
 
   for (int cnt = 0; cnt < argc; ++cnt)
     {
       ENTRY entry = { .key = argv[cnt], .data = &argv[cnt] };
       if (hsearch (entry, ENTER) == NULL)
-	error (EXIT_FAILURE, errno,
-	       gettext ("cannot insert into hash table"));
+	error_exit (errno, _("cannot insert into hash table"));
     }
 
   arlib_init ();
@@ -940,7 +951,7 @@ do_oper_delete (const char *arfname, char **argv, int argc,
 	  ENTRY entry;
 	  entry.key = arhdr->ar_name;
 	  ENTRY *res = hsearch (entry, FIND);
-	  if (res != NULL && (instance < 0 || instance-- == 0)
+	  if (res != NULL && (instance < 0 || --instance == 0)
 	      && !found[(char **) res->data - argv])
 	    found[(char **) res->data - argv] = do_delete = true;
 	}
@@ -1002,7 +1013,7 @@ do_oper_delete (const char *arfname, char **argv, int argc,
       if (newfd != -1)
 	close (newfd);
     nonew:
-      error (0, errno, gettext ("cannot create new file"));
+      error (0, errno, _("cannot create new file"));
       status = 1;
       goto errout;
     }
@@ -1042,22 +1053,23 @@ do_oper_delete (const char *arfname, char **argv, int argc,
     }
 
   /* Set the mode of the new file to the same values the original file
-     has.  */
+     has.  Never complain about fchown failing.  But do it before
+     setting the mode (which might be reset/ignored if the owner is
+     wrong.  */
+  if (fchown (newfd, st.st_uid, st.st_gid) != 0) { ; }
   if (fchmod (newfd, st.st_mode & ALLPERMS) != 0
-      /* Never complain about fchown failing.  */
-      || (({asm ("" :: "r" (fchown (newfd, st.st_uid, st.st_gid))); }),
-	  close (newfd) != 0)
-      || (newfd = -1, rename (tmpfname, arfname) != 0))
+      || close (newfd) != 0)
+    goto nonew_unlink;
+  newfd = -1;
+  if (rename (tmpfname, arfname) != 0)
     goto nonew_unlink;
 
  errout:
-#ifdef DEBUG
   elf_end (elf);
 
   arlib_fini ();
 
   close (fd);
-#endif
 
   not_found (argc, argv, found);
 
@@ -1065,12 +1077,18 @@ do_oper_delete (const char *arfname, char **argv, int argc,
 }
 
 
-static void
+/* Prints the given value in the given buffer without a trailing zero char.
+   Returns false if the given value doesn't fit in the given buffer.  */
+static bool
 no0print (bool ofmt, char *buf, int bufsize, long int val)
 {
   char tmpbuf[bufsize + 1];
-  snprintf (tmpbuf, sizeof (tmpbuf), ofmt ? "%-*lo" : "%-*ld", bufsize, val);
+  int ret = snprintf (tmpbuf, sizeof (tmpbuf), ofmt ? "%-*lo" : "%-*ld",
+		      bufsize, val);
+  if (ret >= (int) sizeof (tmpbuf))
+    return false;
   memcpy (buf, tmpbuf, bufsize);
+  return true;
 }
 
 
@@ -1079,7 +1097,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 		const char *member)
 {
   int status = 0;
-  Elf *elf;
+  Elf *elf = NULL;
   struct stat st;
   int fd = open_archive (arfname, O_RDONLY, 0, &elf, &st, oper != oper_move);
 
@@ -1097,7 +1115,8 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
   if (fd == -1)
     {
       if (!suppress_create_msg)
-	fprintf (stderr, "%s: creating %s\n", AR, arfname);
+	fprintf (stderr, "%s: creating %s\n",
+		 program_invocation_short_name, arfname);
 
       goto no_old;
     }
@@ -1109,16 +1128,15 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
   if (oper != oper_qappend)
     {
       if (hcreate (2 * argc) == 0)
-	error (EXIT_FAILURE, errno, gettext ("cannot create hash table"));
+	error_exit (errno, _("cannot create hash table"));
 
       for (int cnt = 0; cnt < argc; ++cnt)
 	{
 	  ENTRY entry;
-	  entry.key = full_path ? argv[cnt] : basename (argv[cnt]);
+	  entry.key = full_path ? argv[cnt] : (char*)xbasename (argv[cnt]);
 	  entry.data = &argv[cnt];
 	  if (hsearch (entry, ENTER) == NULL)
-	    error (EXIT_FAILURE, errno,
-		   gettext ("cannot insert into hash table"));
+	    error_exit (errno, _("cannot insert into hash table"));
 	}
     }
 
@@ -1192,7 +1210,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
     next:
       cmd = elf_next (subelf);
       if (elf_end (subelf) != 0)
-	error (EXIT_FAILURE, 0, "%s: %s", arfname, elf_errmsg (-1));
+	error_exit (0, "%s: %s", arfname, elf_errmsg (-1));
     }
 
   if (oper != oper_qappend)
@@ -1200,8 +1218,8 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 
  no_old:
   if (member != NULL)
-    error (EXIT_FAILURE, 0, gettext ("position member %s not found"),
-	   member);
+    error_exit (0, _("position member %s not found"),
+		member);
 
   if (oper == oper_move)
     {
@@ -1210,8 +1228,8 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 	{
 	  if (found[cnt] == NULL)
 	    {
-	      fprintf (stderr, gettext ("%s: no entry %s in archive!\n"),
-		       AR, argv[cnt]);
+	      fprintf (stderr, _("%s: no entry %s in archive!\n"),
+		       program_invocation_short_name, argv[cnt]);
 	      status = 1;
 	    }
 
@@ -1224,7 +1242,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
       /* Open all the new files, get their sizes and add all symbols.  */
       for (int cnt = 0; cnt < argc; ++cnt)
 	{
-	  const char *bname = basename (argv[cnt]);
+	  const char *bname = xbasename (argv[cnt]);
 	  size_t bnamelen = strlen (bname);
 	  if (found[cnt] == NULL)
 	    {
@@ -1239,18 +1257,18 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 	  int newfd = open (argv[cnt], O_RDONLY);
 	  if (newfd == -1)
 	    {
-	      error (0, errno, gettext ("cannot open %s"), argv[cnt]);
+	      error (0, errno, _("cannot open %s"), argv[cnt]);
 	      status = 1;
 	    }
 	  else if (fstat (newfd, &newst) == -1)
 	    {
-	      error (0, errno, gettext ("cannot stat %s"), argv[cnt]);
+	      error (0, errno, _("cannot stat %s"), argv[cnt]);
 	      close (newfd);
 	      status = 1;
 	    }
 	  else if (!S_ISREG (newst.st_mode))
 	    {
-	      error (0, errno, gettext ("%s is no regular file"), argv[cnt]);
+	      error (0, errno, _("%s is no regular file"), argv[cnt]);
 	      close (newfd);
 	      status = 1;
 	    }
@@ -1263,7 +1281,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 		   == NULL)
 	    {
 	      fprintf (stderr,
-		       gettext ("cannot get ELF descriptor for %s: %s\n"),
+		       _("cannot get ELF descriptor for %s: %s\n"),
 		       argv[cnt], elf_errmsg (-1));
 	      status = 1;
 	    }
@@ -1274,16 +1292,17 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 			found[cnt]->old_off == -1l ? 'a' : 'r', argv[cnt]);
 
 	      found[cnt]->elf = newelf;
-	      found[cnt]->sec = newst.st_mtime;
-	      found[cnt]->uid = newst.st_uid;
-	      found[cnt]->gid = newst.st_gid;
+	      found[cnt]->sec = arlib_deterministic_output ? 0 : newst.st_mtime;
+	      found[cnt]->uid = arlib_deterministic_output ? 0 : newst.st_uid;
+	      found[cnt]->gid = arlib_deterministic_output ? 0 : newst.st_gid;
 	      found[cnt]->mode = newst.st_mode;
 	      found[cnt]->name = bname;
 
 	      found[cnt]->mem = elf_rawfile (newelf, &found[cnt]->size);
-	      if (found[cnt] == NULL || elf_cntl (newelf, ELF_C_FDDONE) != 0)
-		error (EXIT_FAILURE, 0, gettext ("cannot read %s: %s"),
-		       argv[cnt], elf_errmsg (-1));
+	      if (found[cnt]->mem == NULL
+		  || elf_cntl (newelf, ELF_C_FDDONE) != 0)
+		error_exit (0, _("cannot read %s: %s"),
+			    argv[cnt], elf_errmsg (-1));
 
 	      close (newfd);
 
@@ -1296,13 +1315,11 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 
   if (status != 0)
     {
-#ifdef DEBUG
       elf_end (elf);
 
       arlib_fini ();
 
       close (fd);
-#endif
 
       return status;
     }
@@ -1353,7 +1370,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 		|| (arhdr = elf_getarhdr (subelf)) == NULL)
 	      /* This should never happen since we already looked at the
 		 archive content.  But who knows...  */
-	      error (EXIT_FAILURE, 0, "%s: %s", arfname, elf_errmsg (-1));
+	      error_exit (0, "%s: %s", arfname, elf_errmsg (-1));
 
 	    arlib_add_symbols (subelf, arfname, arhdr->ar_name, cur_off);
 
@@ -1399,7 +1416,7 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 	    close (newfd);
 	}
     nonew:
-      error (0, errno, gettext ("cannot create new file"));
+      error (0, errno, _("cannot create new file"));
       status = 1;
       goto errout;
     }
@@ -1437,7 +1454,11 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 
 	      /* Create the header.  */
 	      struct ar_hdr arhdr;
-	      char tmpbuf[sizeof (arhdr.ar_name) + 1];
+	      /* The ar_name is not actually zero terminated, but we
+		 need that for snprintf.  Also if the name is too
+		 long, then the string starts with '/' plus an index
+		 off number (decimal).  */
+	      char tmpbuf[sizeof (arhdr.ar_name) + 2];
 	      if (all->long_name_off == -1)
 		{
 		  size_t namelen = strlen (all->name);
@@ -1447,19 +1468,41 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 		}
 	      else
 		{
-		  snprintf (tmpbuf, sizeof (arhdr.ar_name) + 1, "/%-*ld",
+		  snprintf (tmpbuf, sizeof (tmpbuf), "/%-*ld",
 			    (int) sizeof (arhdr.ar_name), all->long_name_off);
 		  memcpy (arhdr.ar_name, tmpbuf, sizeof (arhdr.ar_name));
 		}
 
-	      no0print (false, arhdr.ar_date, sizeof (arhdr.ar_date),
-			all->sec);
-	      no0print (false, arhdr.ar_uid, sizeof (arhdr.ar_uid), all->uid);
-	      no0print (false, arhdr.ar_gid, sizeof (arhdr.ar_gid), all->gid);
-	      no0print (true, arhdr.ar_mode, sizeof (arhdr.ar_mode),
-			all->mode);
-	      no0print (false, arhdr.ar_size, sizeof (arhdr.ar_size),
-			all->size);
+	      if (! no0print (false, arhdr.ar_date, sizeof (arhdr.ar_date),
+			      all->sec))
+		{
+		  error (0, errno, _("cannot represent ar_date"));
+		  goto nonew_unlink;
+		}
+	      if (! no0print (false, arhdr.ar_uid, sizeof (arhdr.ar_uid),
+			      all->uid))
+		{
+		  error (0, errno, _("cannot represent ar_uid"));
+		  goto nonew_unlink;
+		}
+	      if (! no0print (false, arhdr.ar_gid, sizeof (arhdr.ar_gid),
+			      all->gid))
+		{
+		  error (0, errno, _("cannot represent ar_gid"));
+		  goto nonew_unlink;
+		}
+	      if (! no0print (true, arhdr.ar_mode, sizeof (arhdr.ar_mode),
+			all->mode))
+		{
+		  error (0, errno, _("cannot represent ar_mode"));
+		  goto nonew_unlink;
+		}
+	      if (! no0print (false, arhdr.ar_size, sizeof (arhdr.ar_size),
+			all->size))
+		{
+		  error (0, errno, _("cannot represent ar_size"));
+		  goto nonew_unlink;
+		}
 	      memcpy (arhdr.ar_fmag, ARFMAG, sizeof (arhdr.ar_fmag));
 
 	      if (unlikely (write_retry (newfd, &arhdr, sizeof (arhdr))
@@ -1494,22 +1537,36 @@ do_oper_insert (int oper, const char *arfname, char **argv, int argc,
 
   /* Set the mode of the new file to the same values the original file
      has.  */
-  if (fd != -1
-      && (fchmod (newfd, st.st_mode & ALLPERMS) != 0
-	  /* Never complain about fchown failing.  */
-	  || (({asm ("" :: "r" (fchown (newfd, st.st_uid, st.st_gid))); }),
-	      close (newfd) != 0)
-	  || (newfd = -1, rename (tmpfname, arfname) != 0)))
-      goto nonew_unlink;
+  if (fd != -1)
+    {
+      /* Never complain about fchown failing.  But do it before
+	 setting the modes, or they might be reset/ignored if the
+	 owner is wrong.  */
+      if (fchown (newfd, st.st_uid, st.st_gid) != 0) { ; }
+      if (fchmod (newfd, st.st_mode & ALLPERMS) != 0
+	  || close (newfd) != 0)
+        goto nonew_unlink;
+      newfd = -1;
+      if (rename (tmpfname, arfname) != 0)
+	goto nonew_unlink;
+    }
 
  errout:
-#ifdef DEBUG
+  for (int cnt = 0; cnt < argc; ++cnt)
+    elf_end (found[cnt]->elf);
+
   elf_end (elf);
 
   arlib_fini ();
 
-  close (fd);
-#endif
+  if (fd != -1)
+    close (fd);
+
+  if (newfd != -1)
+    close (newfd);
 
   return status;
 }
+
+
+#include "debugpred.h"
